import os

import qrcode


class AssetQRCode(qrcode.QRCode):
    def __init__(self):
        super().__init__()

    def make_new_asset_qr(self, data) -> qrcode.image:
        self.clear()
        self.add_data(data)
        return self.make_image()

    def make_new_model_qr(self, data) -> qrcode.image:
        self.clear()
        self.add_data(data)
        return self.make_image()

    def save(self, file_name, data):
        desktop = os.path.expanduser('~/Desktop')
        full_path = os.path.join(desktop, '{}.png'.format(file_name))
        self.prevent_duplicate_file(full_path)
        with open(full_path, 'wb') as fp:
            return self.make_new_asset_qr(data).save(fp)

    def display(self, data):
        img = self.make_new_asset_qr(data)
        img.show()

    def prevent_duplicate_file(self, file):
        if os.path.exists(file):
            raise FileExistsError



DELAY = '~d'
TAB = '~t'
ENTER = '~e'
SPACE = '\x20'


def textify(value: str) -> str:
    contents = [value, TAB]
    return ''.join([delayify(item) for item in contents if item is not None])


def listify(value: str) -> str:
    contents = [SPACE, value, ENTER, TAB]
    return ''.join([delayify(item) for item in contents])


def delayify(value: str, amount=1) -> str:
    delay_count = DELAY * amount
    return delay_count + value


def tabify(amount=1) -> str:
    tab_count = TAB * amount
    return tab_count



from inqry.system_specs import _human_readable
from inqry.system_specs import cfgutil


def create_device_from_device_properties(properties):
    return DeviceSpecs(properties)


def create_devices_from_cfgutil_output():
    return [create_device_from_device_properties(properties) for properties in
            cfgutil.get_hardware_properties_for_attached_devices()]


class DeviceSpecs:
    def __init__(self, device_properties):
        self.serial_number = device_properties.get('serialNumber')
        self.model_identifier = device_properties.get('deviceType')
        self.mobile_storage = _human_readable(device_properties.get('totalDiskCapacity'))
        self.imei = device_properties.get('IMEI')
        self.os_type = 'iOS'
        self.processor_speed = None
        self.processor_name = None
        self.model_name = None
        self.memory = None
        self.drive1 = None
        self.drive2 = None
        self.drive3 = None
        self.drive4 = None

    def get_all_device_components(self):
        return {'Model Identifier': self.model_identifier,
                'Manufacturer': 'Apple',
                'Serial Number (system)': self.serial_number,
                'IMEI': self.imei,
                'Storage': self.mobile_storage}


def get_hardware_overview():
    return create_devices_from_cfgutil_output()



import re
import shlex
import subprocess

BASE_COMMAND = '/usr/sbin/diskutil'


def get_all_physical_ids(diskutil_list_output=None):
    diskutil_list_output = diskutil_list_output or list_all()
    pattern = re.compile(r'(/dev/disk\d+)')
    return re.findall(pattern, diskutil_list_output)


def get_disk_info(disk_id):
    return _get_output_of_diskutil_command(arguments='info {}'.format(disk_id))


def list_all():
    return _get_output_of_diskutil_command(arguments='list')


def _get_output_of_diskutil_command(arguments=None):
    arguments = arguments or ''
    full_command = shlex.split(' '.join([BASE_COMMAND, arguments]))
    return subprocess.check_output(full_command).decode('utf-8')



from plistlib import Plist

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

__version__ = '1.3.2'
__author__ = "Microsoft Apex Lab"
__copyright__ = 'Copyright © 2017 Microsoft. All rights reserved.'
__credits__ = ["Eric Hanko", "Jacob Zaval", "Michael Brown", "Andre Shields", "Ryan Dominguez", "Eammon Hanlon"]
__license__ = 'MIT'
__email__ = 'apxlab@microsoft.com'

__description__ = '''
InQRy is a cross-platform application that generates a single QR code containing the machine's hardware specifications.
This application is designed primarily to be used during a physical inventory procedure.

The QR code contains detailed information about the client machine or device, which can then be scanned it quickly add
assets into a Snipe-IT database.
'''

try:
    plist = Plist.fromFile('Info.plist')
    plist.update(dict(
            CFBundleVersion=__version__,
            CFBundleShortVersionString=__version__,
            CFBundleName='InQRy',
            NSHumanReadableCopyright=__copyright__))
except FileNotFoundError:
    plist = None

setup(
        name='inqry',
        app=['inqry/__main__.py'],
        author=[__author__],
        author_email=__email__,
        description='A cross-platform utility used to generate a QR code containing hardware specs',
        license=__license__,
        long_description=__description__,
        packages=['inqry', 'inqry.system_specs'],
        install_requires=['Pillow',
                          'pytest',
                          'PyYAML',
                          'qrcode',
                          'wmi;platform_system=="Windows"',
                          'pypiwin32;platform_system=="Windows"'],
        tests_require=['pytest'],
        url='https://github.com/Microsoft/InQRy',
        version=__version__,
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Intended Audience :: Information Technology',
            'Intended Audience :: System Administrators',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3 :: Only'],
        options=dict(
                py2app=dict(
                        plist=plist)))



import re
import subprocess
import sys
from tkinter import Button, E, Entry, Label, OptionMenu, StringVar, Tk, W, messagebox

from inqry.asset_qrcode import AssetQRCode
from inqry.form_instructions import FormInstructions
from inqry.system_specs.systemspecs import SystemSpecs


class InQRyGUI:  # TODO: Extract GUI attributes to methods
    def __init__(self):
        self.specs = SystemSpecs()
        self.form_instructions = FormInstructions(self.specs)
        self.asset_qr = AssetQRCode()

        self.root_window = Tk()
        self.root_window.title('InQRy')

        self.form_options_label = Label(self.root_window, text='Form Factor:')
        self.form_options_label.grid(row=1, column=1, sticky=E)

        self.form_options = tuple(self.form_instructions.form_types.keys())[0:2]
        self.form_selection = StringVar()
        self.form_selection.set(self.form_options[0])
        self.form_menu = OptionMenu(self.root_window, self.form_selection, *self.form_options)
        self.form_menu.grid(row=1, column=2, sticky=W)

        self.qrcode_options_label = Label(self.root_window, text='QR Code Type:')
        self.qrcode_options_label.grid(row=2, column=1, sticky=E)

        self.qrcode_options = ['Create Asset']
        self.qrcode_selection = StringVar()
        self.qrcode_selection.set(self.qrcode_options[0])
        self.qrcode_menu = OptionMenu(self.root_window, self.qrcode_selection, *self.qrcode_options)
        self.qrcode_menu.grid(row=2, column=2, sticky=W)

        self.alias_label = Label(self.root_window, text='Alias:')
        self.alias_label.grid(row=3, column=1, sticky=E)

        self.alias_entry = Entry(self.root_window)
        self.alias_entry.grid(row=3, column=2)

        self.asset_tag_label = Label(self.root_window, text='Asset Tag:')
        self.asset_tag_label.grid(row=4, column=1, sticky=E)

        self.asset_tag_entry = Entry(self.root_window)
        self.asset_tag_entry.grid(row=4, column=2)

        # Don't instantiate the "Show" button during inventory
        # self.generate_qr_button = Button(self.root_window, text='Show QR Code', command=self.display)
        # self.generate_qr_button.grid(row=5, column=1)

        self.generate_qr_button = Button(self.root_window, text='Save QR Code to Desktop', command=self.save)
        self.generate_qr_button.grid(row=5, column=2)

    def save(self):
        data = self.gather_user_input()
        filename = data[2] + '-' + data[1]
        try:
            return self.asset_qr.save(filename, self.form_instructions.gui_helper(*data))
        except TypeError:
            print('Improper formatting.')
        except FileExistsError:
            error_message_box('File already exists.')

    def display(self):
        data = self.gather_user_input()
        return self.asset_qr.display(self.form_instructions.gui_helper(*data))

    def gather_user_input(self) -> tuple:
        return self.qrcode_selection.get(), self.get_asset_tag(), self.get_alias(), self.form_selection.get()

    def validate(self, contents, field):
        patterns = {'Alias': re.compile(r'^(v\-)?[A-Za-z]+$'),
                    'Asset Tag': re.compile(r'^E?\d{7}$')}
        return contents if bool(re.match(patterns[field], contents)) else error_message_box(
                '{} is not properly formatted.'.format(field))

    def get_alias(self):
        return self.validate(self.alias_entry.get(), 'Alias')

    def get_asset_tag(self):
        return self.validate(self.asset_tag_entry.get(), 'Asset Tag')

    def obtain_default_dimensions_for_the_root_gui_object(self):
        return tuple(int(_) for _ in self.root_window.geometry().split('+')[0].split('x'))

    def calculate_center_coordinates(self, screen_dimension, current_dimension):
        return (screen_dimension // 2) - (current_dimension // 2)

    def center(self):
        self.root_window.update_idletasks()
        size = self.obtain_default_dimensions_for_the_root_gui_object()
        height = size[0]
        width = size[1]

        x = self.calculate_center_coordinates(self.root_window.winfo_screenwidth(), height)
        y = self.calculate_center_coordinates(self.root_window.winfo_screenheight(), width)
        self.root_window.geometry('{}x{}+{}+{}'.format(height, width, x, y))


def error_message_box(message: str):
    messagebox.showerror('Oops!', message)


def mobile_capability():  # TODO: Re-implement mobile_capability() into GUI
    if sys.platform == 'darwin':
        try:
            subprocess.check_output(['/usr/local/bin/cfgutil'])
            return 'active'
        except FileNotFoundError:
            print(
                    '''
                    No such file or directory: "/usr/local/bin/cfgutil"

                    You must install cfgutil using Apple Configurator in order to
                    use InQRy with a mobile device.
                    ''')
            return 'disable'
    else:
        pass



import sys
from inqry.system_specs.systemspecs import SystemSpecs
from inqry import barcode


class FormInstructions:
    def __init__(self, specs: SystemSpecs):
        self.processor = '{} {}'.format(specs.processor_speed, specs.processor_name)
        self.form_types = {'Desktop': [self.processor, specs.memory,
                                       specs.drive1, specs.drive2,
                                       specs.drive3, specs.drive4],
                           'Portable': [self.processor, specs.memory,
                                        specs.drive1]
                            }
        self.model_identifier = specs.model_identifier
        self.model_name = specs.model_name
        self.serial_number = specs.serial_number

    def get_asset_sequence(self, form_type) -> str:
        try:
            asset_data = self.form_types[form_type]
            return ''.join([barcode.textify(component) for component in asset_data])
        except TypeError('Form type is required.'):
            sys.exit(1)

    def get_user_sequence(self, user) -> str:
        try:
            user_sequence = [barcode.SPACE, user, barcode.ENTER, barcode.TAB]
            return ''.join([barcode.delayify(char) for char in user_sequence])
        except TypeError('User entry is required.'):
            sys.exit(1)

    def new_asset(self, asset_tag, user, form_type) -> str:
        status = 'Ready'
        return barcode.textify(asset_tag) + barcode.listify(self.model_identifier) + self.get_asset_sequence(
            form_type) + barcode.listify(status) + self.get_user_sequence(user) + barcode.delayify(
            self.serial_number)

    def new_model(self) -> str:
        return barcode.textify(self.model_name) + barcode.tabify(2) + barcode.delayify(
            self.model_identifier)

    def gui_helper(self, qrcode_type, *args) -> str:
        qrcode_types = {'Create Asset': self.new_asset(*args),
                        'New Model': self.new_model()}
        return qrcode_types[qrcode_type]



import re
from inqry.system_specs import win_physical_disk


def get_all_physical_disks_from_friendly_names():
    friendly_names = win_physical_disk.get_physical_disk_friendly_names()
    return [Disk(win_physical_disk.get_disk_info_from_friendly_name(friendly_name))
            for friendly_name in friendly_names]


def get_internal_storage():
    internal_disks = get_all_physical_disks_from_friendly_names()
    return [disk for disk in internal_disks if disk.is_internal]


class Disk(object):
    def __init__(self, windows_disk):
        self.windows_disk = windows_disk

    @property
    def bustype(self):
        pattern = re.compile(r'(?:BusType\s+:\s)([a-zA-Z]+)')
        return re.findall(pattern, self.windows_disk)[0]

    @property
    def device_location(self):
        return 'Internal' if self.is_internal else 'External'

    @property
    def device_name(self):
        pattern = re.compile(r'(?:FriendlyName\s+:\s)([\w\d\-._ ]+)')
        return re.findall(pattern, self.windows_disk)[0]

    @property
    def is_internal(self):
        valid_bus_types = ['SATA', 'RAID', 'NVMe', 'ATA']
        return self.bustype in valid_bus_types

    @property
    def is_external(self):
        return self.is_internal is False

    @property
    def type(self):
        pattern = re.compile(r'(?:MediaType\s+:\s)([A-Z]+)')
        return re.findall(pattern, self.windows_disk)[0]

    @property
    def is_ssd(self):
        return self.type == 'SSD'

    @property
    def size(self):
        pattern = re.compile(r'(?:Size\s+:\s)([\d]+)')
        return self._human_readable(re.findall(pattern, self.windows_disk)[0])

    @staticmethod
    def _human_readable(component):
        size = round(int(component) / 10 ** 9)
        if size >= 1000:
            return str(size / 1000) + " TB"
        else:
            return str(size) + " GB"
